@@ -11,24 +11,25 @@
 import re
 
 from docutils import nodes
-from docutils.parsers.rst import directives
+from docutils.parsers.rst import directives, Directive
 
 from sphinx import addnodes
 from sphinx.roles import XRefRole
-from sphinx.locale import l_, _
+from sphinx.locale import _
 from sphinx.domains import Domain, ObjType, Index
 from sphinx.directives import ObjectDescription
 from sphinx.util.nodes import make_refnode
-from sphinx.util.compat import Directive
 from sphinx.util.docfields import Field, GroupedField, TypedField
-
+from sphinx import __version__ as sphinx_version
 
 php_sig_re = re.compile(
-    r'''^ ([\w.]*\:\:)?          # class name(s)
-          (\$?\w+)  \s*          # thing name
-          (?: \((.*)\)           # optional: arguments
-          (?:\s* -> \s* (.*))?   # return annotation
-          )? $                   # and nothing more
+    r'''^ (public\ |protected\ |private\ )? # visibility
+          (final\ |abstract\ |static\ )?    # modifiers
+          ([\w.]*\:\:)?                     # class name(s)
+          (\$?\w+)  \s*                     # thing name
+          (?: \((.*)\)                      # optional: arguments
+          (?:\s* -> \s* (.*))?              # return annotation
+          )? $                              # and nothing more
           ''', re.VERBOSE)
 
 
@@ -115,19 +116,19 @@ class PhpObject(ObjectDescription):
     }
 
     doc_field_types = [
-        TypedField('parameter', label=l_('Parameters'),
+        TypedField('parameter', label=_('Parameters'),
                    names=('param', 'parameter', 'arg', 'argument'),
                    typerolename='obj', typenames=('paramtype', 'type')),
-        TypedField('variable', label=l_('Variables'), rolename='obj',
+        TypedField('variable', label=_('Variables'), rolename='obj',
                    names=('var', 'ivar', 'cvar'),
                    typerolename='obj', typenames=('vartype',)),
-        GroupedField('exceptions', label=l_('Throws'), rolename='exc',
+        GroupedField('exceptions', label=_('Throws'), rolename='exc',
                      names=('throws', 'throw', 'exception', 'except'),
                      can_collapse=True),
-        Field('returnvalue', label=l_('Returns'), has_arg=False,
+        Field('returnvalue', label=_('Returns'), has_arg=False,
               names=('returns', 'return')),
-        Field('returntype', label=l_('Return type'), has_arg=False,
-              names=('rtype', 'returntype')),
+        Field('returntype', label=_('Return type'), has_arg=False,
+              names=('rtype', 'returntype'), bodyrolename='obj'),
     ]
 
     def get_signature_prefix(self, sig):
@@ -156,7 +157,7 @@ class PhpObject(ObjectDescription):
         if m is None:
             raise ValueError
 
-        name_prefix, name, arglist, retann = m.groups()
+        visibility, modifiers, name_prefix, name, arglist, retann = m.groups()
 
         if not name_prefix:
             name_prefix = ""
@@ -203,8 +204,14 @@ class PhpObject(ObjectDescription):
         signode['class'] = self.class_name = classname
         signode['fullname'] = fullname
 
+        if visibility:
+            signode += addnodes.desc_annotation(visibility, visibility)
+
         sig_prefix = self.get_signature_prefix(sig)
 
+        if modifiers and not (sig_prefix and 'static' in sig_prefix):
+                signode += addnodes.desc_annotation(modifiers, modifiers)
+
         if sig_prefix:
             signode += addnodes.desc_annotation(sig_prefix, sig_prefix)
 
@@ -274,12 +281,11 @@ class PhpObject(ObjectDescription):
             self.state.document.note_explicit_target(signode)
             objects = self.env.domaindata['php']['objects']
             if fullname in objects:
-                self.env.warn(
-                    self.env.docname,
+                self.state_machine.reporter.warning(
                     'duplicate object description of %s, ' % fullname +
                     'other instance in ' +
                     self.env.doc2path(objects[fullname][0]),
-                    self.lineno)
+                    line=self.lineno)
             objects[fullname] = (self.env.docname, self.objtype)
 
         indextext = self.get_index_text(modname, name_cls)
@@ -451,6 +457,28 @@ class PhpNamespace(Directive):
         return ret
 
 
+class PhpCurrentNamespace(Directive):
+    """
+    This directive is just to tell Sphinx that we're documenting
+    stuff in namespace foo, but links to namespace foo won't lead here.
+    """
+
+    has_content = False
+    required_arguments = 1
+    optional_arguments = 0
+    final_argument_whitespace = False
+    option_spec = {}
+
+    def run(self):
+        env = self.state.document.settings.env
+        modname = self.arguments[0].strip()
+        if modname == 'None':
+            env.temp_data['php:namespace'] = None
+        else:
+            env.temp_data['php:namespace'] = modname
+        return []
+
+
 class PhpXRefRole(XRefRole):
     """
     Provides cross reference links for PHP objects
@@ -480,8 +508,8 @@ class PhpNamespaceIndex(Index):
     """
 
     name = 'modindex'
-    localname = l_('PHP Namespace Index')
-    shortname = l_('namespaces')
+    localname = _('PHP Namespace Index')
+    shortname = _('namespaces')
 
     def generate(self, docnames=None):
         content = {}
@@ -548,16 +576,16 @@ class PhpDomain(Domain):
     name = 'php'
     label = 'PHP'
     object_types = {
-        'function': ObjType(l_('function'), 'func', 'obj'),
-        'global': ObjType(l_('global variable'), 'global', 'obj'),
-        'const': ObjType(l_('const'), 'const', 'obj'),
-        'method': ObjType(l_('method'), 'meth', 'obj'),
-        'class': ObjType(l_('class'), 'class', 'obj'),
-        'attr': ObjType(l_('attribute'), 'attr', 'obj'),
-        'exception': ObjType(l_('exception'), 'exc', 'obj'),
-        'namespace': ObjType(l_('namespace'), 'ns', 'obj'),
-        'interface': ObjType(l_('interface'), 'interface', 'obj'),
-        'trait': ObjType(l_('trait'), 'trait', 'obj'),
+        'function': ObjType(_('function'), 'func', 'obj'),
+        'global': ObjType(_('global variable'), 'global', 'obj'),
+        'const': ObjType(_('const'), 'const', 'obj'),
+        'method': ObjType(_('method'), 'meth', 'obj'),
+        'class': ObjType(_('class'), 'class', 'obj'),
+        'attr': ObjType(_('attribute'), 'attr', 'obj'),
+        'exception': ObjType(_('exception'), 'exc', 'obj'),
+        'namespace': ObjType(_('namespace'), 'ns', 'obj'),
+        'interface': ObjType(_('interface'), 'interface', 'obj'),
+        'trait': ObjType(_('trait'), 'trait', 'obj'),
     }
 
     directives = {
@@ -572,6 +600,8 @@ class PhpDomain(Domain):
         'interface': PhpClasslike,
         'trait': PhpClasslike,
         'namespace': PhpNamespace,
+        'currentmodule': PhpCurrentNamespace,
+        'currentnamespace': PhpCurrentNamespace,
     }
 
     roles = {
@@ -597,20 +627,37 @@ class PhpDomain(Domain):
     ]
 
     def clear_doc(self, docname):
-        for fullname, (fn, _) in self.data['objects'].items():
+        for fullname, (fn, _l) in list(self.data['objects'].items()):
             if fn == docname:
                 del self.data['objects'][fullname]
-        for ns, (fn, _, _) in self.data['namespaces'].items():
+        for ns, (fn, _x, _x) in list(self.data['namespaces'].items()):
             if fn == docname:
                 del self.data['namespaces'][ns]
 
+    def merge_domaindata(self, docnames, otherdata):
+        for fullname, (fn, objtype) in otherdata['objects'].items():
+            if fn in docnames:
+                self.data['objects'][fullname] = (fn, objtype)
+        for namespace, data in otherdata['namespaces'].items():
+            if data[0] in docnames:
+                self.data['namespaces'][namespace] = data
+
+    def resolve_any_xref(self, env, fromdocname, builder,
+                         target, node, contnode):
+        for typ in self.roles:
+            resolve = self.resolve_xref(env, fromdocname, builder,
+                                        typ, target, node, contnode)
+            if resolve:
+                return [('php:%s' % typ, resolve)]
+        return []
+
     def resolve_xref(self, env, fromdocname, builder,
                      typ, target, node, contnode):
         if (typ == 'ns' or
                 typ == 'obj' and target in self.data['namespaces']):
             docname, synopsis, deprecated = self.data['namespaces'].get(
                 target,
-                ('', '', '', '')
+                ('', '', '')
             )
             if not docname:
                 return None
@@ -697,3 +744,5 @@ class PhpDomain(Domain):
 
 def setup(app):
     app.add_domain(PhpDomain)
+
+    return {'version': sphinx_version, 'parallel_read_safe': True}
